package ims.trans;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.IOracle;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.ParametersFloat.IUpdStruct;
import ims.trans.Options.UpdateType;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.util.TRSeqPathAnalysis;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.IOUtils;
import ims.util.Util;

public abstract class AbstractTrainer implements ITrainer {

	protected abstract int doIteration(int itemCount,List<Instance> instances);
	
	protected final Options options;
	protected final SymbolTable symTab;
	protected final IOracle oracle;
	protected final ITransitionFilter transitionFilter;
	protected final FeatureExtractor featureExtractor;
	protected final Long2IntInterface l2i;
	protected final ParametersFloat parameters;
	protected final UpdateType ut;
	protected final InstanceMapping instanceMapping;
	protected final boolean mergeEquivalentStates;
	protected final LossFunction lossFunction;
	
	public AbstractTrainer(Options options, SymbolTable symTab, IOracle oracle,ITransitionFilter transitionFilter, FeatureExtractor featureExtractor,Long2IntInterface l2i, ParametersFloat parameters,LossFunction lossFunction,InstanceMapping instanceMapping) {
		this.options = options;
		this.symTab = symTab;
		this.oracle = oracle;
		this.transitionFilter = transitionFilter;
		this.featureExtractor = featureExtractor;
		this.l2i = l2i;
		this.parameters = parameters;
		this.mergeEquivalentStates=options.mergeEquivalentStates;
		this.ut=options.ut;
		this.lossFunction=lossFunction;
		this.instanceMapping=instanceMapping;
	}
	
	public static interface Tester {
		public void applyDecoder(AbstractDecoder decoder,SentenceReader<?> reader,File output,boolean reversed) throws IOException;
		public void evalTest(File goldFile,File predFile);
	}
	
	protected void doInterIterTest(Tester tester,int itemCount,int iter) throws IOException, UnsupportedEncodingException, FileNotFoundException {
		if(options.testEveryIter){
			l2i.freeze();
			float[] params;
			if(Options.useAveraging)
				params=parameters.averagedCopy(itemCount);
			else
				params=parameters.parameters;
			AbstractDecoder d = getDecoder(params);
			
			File out;
			if(options.saveEveryOut)
				out=new File(options.out.toString()+".i"+String.format("%03d", iter));
			else
				out=options.out;
			DBO.printNoPrefix("Testing on ");
			tester.applyDecoder(d,ReaderWriterFactory.getReader(options.test,Options.format.sentenceClass,options.reversed),out,options.reversed);
			tester.evalTest(options.gold, out);
			l2i.unFreeze();
		}
	}

	private AbstractDecoder getDecoder(float[] params) {
		AbstractDecoder d;
		if(options.sample){
			
			d=new SamplingDecoder(params,l2i,symTab,transitionFilter,featureExtractor,options.mergeEquivalentStates,options.labeled,instanceMapping,SamplingStoppingCriterion.get(options.stopC, options.sampleK));
		} else {
			d=new Decoder(params, l2i, symTab, transitionFilter, featureExtractor,options.mergeEquivalentStates,options.labeled,instanceMapping,options.beam,Options.goldBeam);	
		}
		return d;
	}

	@Override
	public Model train(Tester tester,IInstanceCreator instanceCreator) throws IOException {
		int itemCount=0;
		for(int iter=0;iter<options.iterations;++iter){
			List<Instance> instances=instanceCreator.getInstances();
			if(Options.shuffleStrategy.shuffleBetweenIterations)
				ArrayShuffler.SINGLETON.shuffle(instances);
			DBO.println("Iteration "+iter+" \t");
			boolean threadSafe=!l2i.frozen();
			itemCount=doIteration(itemCount,instances);
			printNonZero(parameters);
			doInterIterTest(tester,itemCount,iter);
			oracle.nextIter();
			if(Options.outTRFilePF!=null)
				dumpTRSeqs(iter,instances);
			if(options.saveEveryModel && options.model!=null){
				Model m=new Model(featureExtractor, symTab, parameters.averagedCopy(itemCount), l2i, transitionFilter,options.beam,options.mergeEquivalentStates,options.labeled,instanceMapping,options.transitionSystem,options.reversed,options.sampleK,options.stopC);
				File mdl=new File(options.model.toString()+".i"+String.format("%03d", iter));
				Train.saveModel(m, mdl, null);
			}
		}

		if(Options.useAveraging)
			parameters.average(itemCount);
		
		return new Model(featureExtractor, symTab, parameters.parameters, l2i, transitionFilter,options.beam,options.mergeEquivalentStates,options.labeled,instanceMapping,options.transitionSystem,options.reversed,options.sampleK,options.stopC);
	}

	private void dumpTRSeqs(int iter,List<Instance> instances) throws IOException {
		File outFile=TRSeqPathAnalysis.getFileName(Options.outTRFilePF, iter);
		BufferedWriter writer=IOUtils.getBufferedWriter(outFile);
		for(Instance inst:instances){
			List<String> l=new ArrayList<String>();
			for(int i:inst.lastTRSeq)
				l.add(symTab.transition2label[i].getFullTransitionLabel());
			writer.write(inst.sentId+":");
			writer.write(Util.joinStrings(l, ","));
			writer.newLine();
		}
		writer.close();
	}

	private void printNonZero(ParametersFloat parameters) {
		int nz=parameters.countNZ();
		int total=parameters.total.length;
		int l=Integer.toString(total).length();
		String format="Parameters: nz: %"+l+"d  / total: %"+l+"d";
		DBO.println(String.format(format,nz,total));
	}

	protected void bookkeepLastPredTRSeq(State gold, Instance inst, Bookkeeping bookkeeping) {
		int[] newSeq;
		if(gold.isTerminal()){
			newSeq=new int[gold.totalTransitions];
		} else {
			newSeq=new int[gold.totalTransitions+1];
			newSeq[gold.totalTransitions]=-1; //sentinel for marking incomplete sequence (for early update)
		}
		State c=gold;
		for(int i=gold.totalTransitions-1;i>=0;--i){
			newSeq[i]=c.lastAction.getId();
			c=c.previous;
		}
		
		//Check if we have the same (partial) seq as last time
		boolean sameSeqAsLast=true;
		if(inst.lastTRSeq==null){
			sameSeqAsLast=false;
		} else {
			for(int i=0,m=Math.min(newSeq.length, inst.lastTRSeq.length);i<m && inst.lastTRSeq[i]!=-1 && newSeq[i]!=-1;++i){
				if(inst.lastTRSeq[i]!=newSeq[i]){
					sameSeqAsLast=false;
					break;
				}
			}
		}
		
		if(sameSeqAsLast)
			bookkeeping.sameTRSeqAsLast++;
		
		//Check if we have the same (partial) seq as canon
		boolean sameAsCanon;
		if(newSeq.length>inst.transSequence.length){
			sameAsCanon=false;
		} else {
			sameAsCanon=true;
			for(int i=0;i<newSeq.length;++i){
				if(newSeq[i]==-1 && i==newSeq.length-1)
					break;
				if(newSeq[i]!=inst.transSequence[i]){
					sameAsCanon=false;
					break;
				}
			}
		}
		
		if(sameAsCanon){
			bookkeeping.sameTRSeqAsCanon++;
		}
		
		inst.lastTRSeq=newSeq;
	}

	static class Bookkeeping {
		int goldTrans=0;
		int predTrans=0;
		int nullReturns=0;
		int upds=0;
		int updPairs=0;
		int goldAssignedHeads=0;
		int predAssignedHeads=0;
		int goldSwaps=0;
		int predSwaps=0;
		int forkDistSum=0;
		int optGoldLen=0;
		int sameTRSeqAsLast=0;
		int sameTRSeqAsCanon=0;
	}
	
	protected String getTRSeqStats(Bookkeeping bookkeeping,int sentCount) {
		double goldIdenticalToLast=100.0*bookkeeping.sameTRSeqAsLast/sentCount;
		double goldIdenticalToCanon=100.0*bookkeeping.sameTRSeqAsCanon/sentCount;
		return String.format("GSeqLast: %6.2f  |  GSeqCanon: %6.2f", goldIdenticalToLast,goldIdenticalToCanon);
	}
	
	protected boolean checkCorrectTree(State pred, Instance inst) {
		return Arrays.equals(pred.treeSig,inst.head) && (pred.lblSig==null || Arrays.equals(pred.lblSig,inst.lbl));
	}
	
	protected double doUpdate(int itemCount, IFV<FV3> goldFV, IFV<FV3> predFV,Instance inst, State bestPrediction, State correctState) {
		double goldScore=correctState.score();
		double predScore=bestPrediction.score();

		XrecoverFeaturesAndFillVector(goldFV,correctState,inst,true);
		XrecoverFeaturesAndFillVector(predFV,bestPrediction,inst,true);
		double loss=lossFunction.getLoss(correctState,bestPrediction,inst.form.length); //This is oLoss

		if(Options.AUG_LOSS){
			//Since we do inexact search, the augmented loss transitions can have side effects that incur additional loss in absolute terms. <--- Actually, I guess the reason isn't the inexact search but more a consequence of how certain arcs are ruled out by certain other actions... Or is it because the loss function doesn't factor over transitions meaningfully? (Since introducing an arc before a dependent has gathered all its dependents will lead to observed loss, even though this arc transition is augmented with loss all the time?)
			//E.g., we have gold and pred seqs with 10 transitions each. Pred contains 5 transitions that had its loss augmented by 3.5 in total. (this means pred has a score of 3.5 greater than w*fi(pred))
			//      However, these 5 lossaugmented transitions actually cause a greater loss in reality (when we apply the loss function offline), say 5.
			//
			//If tau = (p - g + l) / ( ||...||), where p denotes the actual score of the prediction, i.e., w*fi(pred), then we may be adding a too great loss here if we use what we get from the loss function.
			//
			//We can either let
			// l = oLoss, or
			// l = aLoss
			// where oLoss is "observed"  loss, i.e., what the loss function gives us (5 above), and
			//       aloss in "augmented" loss, i.e., what we have augmented the transitions with (3.5 above)
			double actPredScore=predFV.getScore(parameters.parameters);
			double oLoss=loss;
			double aLoss=predScore-actPredScore; //This is aLoss
			
//			predScore=actPredScore;
//			loss=aLoss;
//			loss=oLoss;
			
			//What's most funny, however, is that it seems like having a bigger numerator when computing tau seems to give better results.
			//I'm guessing the reason is that it reaches better numbers earlier, and that this works out better when doing averaging.
			//Now the real question is whether to just up the numerator of tau, or whether the loss in general should be increased (i.e., so that the loss augmentation makes it even more difficult to get things right during learning)
			//Here we first try by just playing with the loss a bit and do the following two things:
			//
			//1) Let predScore stay the same (with bonus for loss augmentation; this is what we had earlier, before correcting for the augmented loss in predScore)
			//2) Then we take the oLoss and double it (just to make up some number that makes tau bigger)
			
			
			//XXX There's something fishy going on here. It seems like it's useful to just make the numerator for the tau a bit bigger
			//    but I don't understand exactly what's going on here. I should look into this later. 
			
			//The following is my interpretation of how to properly do the lossaugmentation
//			{
//				predScore=actPredScore;
//				loss=oLoss; //This could be aLoss too, that's upp for discussion, given that we don't do exact search
//			}
			//This way was the inital bug (seems better)
//			{
//				predScore=predScore;
//				loss=oLoss;
//			}
			//Just making the numerator even bigger also seems to be better
//			{
//				predScore=predScore;
//				loss=2*oLoss;
//			}
		}
		
		IUpdStruct upd=parameters.getUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
		parameters.update(upd);
		return loss;
	}

	void XrecoverFeaturesAndFillVector(IFV<?> fv, State finalState,Instance inst,boolean clearFv) {
		XrecoverFeaturesAndFillVector(fv,finalState,inst,clearFv,-1);
	}

	void recoverFeaturesAndFillVector(IFV<?> fv, State finalState,Instance inst,boolean clearFv) {
		recoverFeaturesAndFillVector(fv,finalState,inst,clearFv,-1);
	}
	
	/*
	 * XXX NOTE that the skipUntil has a bit of a margin, so it includes one (or two?) previous states too. The main
	 *          idea is just that it saves going all the way if you know the fork point
	 */
	void recoverFeaturesAndFillVector(IFV<?> fv, State finalState,Instance inst,boolean clearFv,int skipUntil) {
		if(clearFv)
			fv.clear();
		ITransition[] transitions=State.getTransitionSequence(finalState);
		
		//Now move forward and extract the features again, and add them to the fv.
		//s is now start state.
		State s=State.getInitialState(inst.head.length,options.labeled,null);
//		long[] sink=new long[FeatureExtractor.SINK_MIN_SIZE];
		long[] sink=featureExtractor.sink;
		for(ITransition tr:transitions){
			if(s.totalTransitions>=skipUntil){
				int f=featureExtractor.extractFeatures(s, sink, inst);
				for(int j=0;j<f;++j){
					sink[j]<<=featureExtractor.bA;
					sink[j]|=tr.getId();
					fv.add(l2i.l2i(sink[j]));
				}
			}
			s=tr.apply(s,0); //We don't care about the score here anyway, so we make it 0
		}
	}
	
	/*
	 * XXX NOTE that the skipUntil has a bit of a margin, so it includes one (or two?) previous states too. The main
	 *          idea is just that it saves going all the way if you know the fork point
	 */
	void XrecoverFeaturesAndFillVector(IFV<?> fv, State finalState,Instance inst,boolean clearFv,int skipUntil) {
		if(clearFv)
			fv.clear();
		
		long[] sink=featureExtractor.sink;
		State c=finalState.previous;
		for(ITransition last=finalState.lastAction;last!=null;last=c.lastAction,c=c.previous){
			if(c.totalTransitions<skipUntil)
				break;
			int f=featureExtractor.extractFeatures(c, sink, inst);
			for(int j=0;j<f;++j){
				sink[j]<<=featureExtractor.bA;
				sink[j]|=last.getId();
				fv.add(l2i.l2i(sink[j]));
			}
			
		}
		
//		ITransition[] transitions=State.getTransitionSequence(finalState);
//		
//		//Now move forward and extract the features again, and add them to the fv.
//		//s is now start state.
//		State s=State.getInitialState(inst.head.length,options.labeled,null);
////		long[] sink=new long[FeatureExtractor.SINK_MIN_SIZE];
//		
//		for(ITransition tr:transitions){
//			if(s.totalTransitions>=skipUntil){
//				int f=featureExtractor.extractFeatures(s, sink, inst);
//				for(int j=0;j<f;++j){
//					sink[j]<<=featureExtractor.bA;
//					sink[j]|=tr.getId();
//					fv.add(l2i.l2i(sink[j]));
//				}
//			}
//			s=tr.apply(s,0); //We don't care about the score here anyway, so we make it 0
//		}
	}

}
