package ims.trans.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import ims.data.ISentence;
import ims.data.Instance;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.AbstractDynamicOracle;
import ims.oracle.Configuration.TR;
import ims.oracle.IConfiguration;
import ims.oracle.MinimalConfiguration;
import ims.oracle.OracleFactory;
import ims.trans.Options;
import ims.trans.Options.TransitionSystem;
import ims.util.IOUtils;
import ims.util.MutableInt;

public class TRSeqPathAnalysis {

	static boolean labeled;
	
	public static void main(String[] args) throws IOException{
		Options options=new Options(args);
		labeled=options.labeled;
		SentenceReader<?> reader=ReaderWriterFactory.getTrainReader(options);
//		AbstractDynamicOracle oracle=OracleFactory.getDynamicOracle(options.transitionSystem);
		AbstractDynamicOracle oracle=OracleFactory.getDynamicOracle(TransitionSystem.ArcSFullDynSwap); //Always use the full thing
		List<Map<Integer,List<String>>> transHistory=readTransHistory(options);
		if(transHistory.isEmpty()){
			System.err.println("No matching files of tr seqs");
		}
		for(Map<Integer,List<String>> m:transHistory){
			countAmbiguities(m,reader,oracle);
		}
	}

	private static void countAmbiguities(Map<Integer, List<String>> m, SentenceReader<?> reader, AbstractDynamicOracle oracle) {
		int totalSwaps=0;
		int i=0;
		Map<String,MutableInt> ambiguities=new TreeMap<>();
		Set<String> amb=new TreeSet<>();
		for(ISentence s:reader){
			int[] heads=s.getGoldHeads();
			List<String> trs=m.get(i);
			boolean treeMatch=sanityCheckMatchingTree(heads,trs);
			if(!treeMatch)
				throw new Error("tree mismatch");
			
			String[] lbls=s.getGoldLabels();
			oracle.initSentence(heads);
			
			MinimalConfiguration minConf=new MinimalConfiguration(heads);
			for(String t:trs){
				List<TR> possible=oracle.getActions(minConf, new Instance(heads));
				TR tr=TR.fromString(t);
				if(!possible.contains(tr))
					throw new Error("mismatch");
				if(tr==TR.SW)
					++totalSwaps;
				if(possible.size()>1)
					recordAmbiguity(heads,lbls,ambiguities,possible,tr,minConf,amb,s);
				minConf.applyTR(tr);
			}
			
			++i;
		}
		
//		MutableInt.printEntryList(ambiguities.entrySet(), 30, System.out);
		printEntries(ambiguities,30,amb);
		System.out.println();
		System.out.printf("Total Swaps: %,d\n",totalSwaps);
		System.out.println("========================================");
	}

	private static void printEntries(Map<String, MutableInt> ambiguities, int stringWidth, Set<String> amb) {
		for(String a:amb){
			TR nonTR=TR.LA;
			if(a.startsWith("{SH,SW"))
				nonTR=TR.SW;
			else if(a.startsWith("{SH,RAS"))
				nonTR=TR.RAS;
			
			String[] ops={a+CHOSEN_DELIM+TR.SH.toString(),a+CHOSEN_DELIM+nonTR.toString()};
			MutableInt[] mis=new MutableInt[]{ambiguities.get(ops[0]),ambiguities.get(ops[1])};
		
			int total=0;
			for(MutableInt mi:mis)
				if(mi!=null)
					total+=mi.getValue();
			double[] perc=new double[2];
			for(int i=0;i<2;++i)
				if(mis[i]!=null)
					perc[i]=100.0*mis[i].getValue()/total;
			
			for(int i=0;i<2;++i)
				if(mis[i]!=null)
					System.out.printf("%-"+stringWidth+"s  %10d  - %6.2f\n", ops[i],mis[i].getValue(),perc[i]);
		}
	}
	
	static final String CHOSEN_DELIM="-+-+-";

	private static void recordAmbiguity(int[] heads, String[] lbls,	Map<String, MutableInt> ambiguities, List<TR> possible, TR tr,IConfiguration conf, Set<String> amb, ISentence s) {
		StringBuilder sb=new StringBuilder("{SH,");
		TR nonShift=possible.get(0);
		if(nonShift==TR.SH)
			nonShift=possible.get(1);
		
		int stackSize=getStackSize(conf);
		String s0tag=s.getGoldPOS()[conf.peekStack(0)];
		
		if(nonShift!=TR.SW && labeled){ //Then it's labeled
			final String label;
			if(nonShift==TR.RAS)
				label=lbls[conf.peekStack(0)];
			else
				label=lbls[conf.peekStack(1)];
			sb.append(nonShift).append("-").append(label);
		} else {
			sb.append(nonShift.toString());
		}
		sb.append("}");
//		sb.append("-").append(stackSize);
		sb.append("-").append(s0tag);
		amb.add(sb.toString());
		
		sb.append(CHOSEN_DELIM).append(tr);
		
		String key=sb.toString();
		MutableInt.incrementMapValue(ambiguities, key);
	}

	private static int getStackSize(IConfiguration conf) {
//		return conf.stackSize(); //TODO bin this
//		return 0;
		int stackSize=conf.stackSize();
		if(stackSize<6)
			return stackSize;
		if(stackSize<=10)
			return 10;
		else
			return 20;
	}

	private static boolean sanityCheckMatchingTree(int[] heads, List<String> trs) {
		MinimalConfiguration minConf=new MinimalConfiguration(heads);
		for(String t:trs){
			TR tr=TR.fromString(t);
			minConf.applyTR(tr);
		}
		if(!minConf.isTerminal() || !Arrays.equals(heads,minConf.getHeads()))
			return false;
		else
			return true;
	}

	private static List<Map<Integer, List<String>>> readTransHistory(Options options) throws IOException {
		if(Options.outTRFilePF==null)
			throw new Error("forgot to set -outTRFilePF");
		List<Map<Integer,List<String>>> r=new ArrayList<>();
		for(int i=0;i<options.iterations;++i){
			File f=getFileName(Options.outTRFilePF,i);
			if(!f.exists())
				break;
			System.out.println("Reading "+f);
			BufferedReader reader=IOUtils.getBufferedReader(f);
			String line;
			Map<Integer,List<String>> m=new TreeMap<>();
			while((line=reader.readLine())!=null){
				String[] a=line.split(":",2);
				Integer sentId=Integer.parseInt(a[0]);
				String[] trSeq=a[1].split(",");
				m.put(sentId, Arrays.asList(trSeq));
			}
			r.add(m);
			reader.close();
		}
		return r;
	}
	
	public static File getFileName(String pf,int iteration){
		return new File(pf+String.format(".i%03d.gz", iteration));
	}
	
}
