package ims.util;

import java.lang.reflect.Array;

/*
 * Note that this heap is not stable, so if you want to mimic proper stable sorting to get the k best items, then you need to ensure that there are NO TIES.
 * If you subclass MinHeap.AbstractTieBreakingHeapItem and assign tieBreakers ascendingly wrt to how items are seen then you should be safe 
 */

public class MinHeap<T extends MinHeap.HasScore & Comparable<T>> {
	public interface HasScore {
		public double score();
	}
	public static abstract class AbstractTieBreakingHeapItem<T extends AbstractTieBreakingHeapItem<T> & HasScore> implements HasScore,Comparable<T> {
		final int tieBreaker;
		protected AbstractTieBreakingHeapItem(int tieBreaker){
			this.tieBreaker=tieBreaker;
		}

		@Override
		public int compareTo(T o) {
			int a=Double.compare(score(), o.score());
			if(a!=0)
				return a;
			return Integer.compare(o.tieBreaker, tieBreaker);
		}
	}
	
	private T[] arr;
	int size=0;

	private int nextTieBreaker=1;
	
	public MinHeap(int max){
		this.arr=makeArray(max);
	}
	
	public MinHeap(T[] arr){
		this.arr=arr;
	}
	public void insert(T t){
		if(size==arr.length){
			popAndInsert(t);
			//Pop and insert
		} else { //Toss
			arr[size++]=t;
			if(size==arr.length)
				heapify();
		}
	}
	public boolean offer(T item) {
		return offer(item.score());
	}
	public boolean offer(double d){
		if(size==arr.length){
			if(d>arr[0].score()){
				return true; //The offered element is greater, then we want it
			} else {
				return false; //The offered element is less, then we don't want it
			}
		} else {
			return true;
		}
	}
	public boolean isEmpty() {
		return size==0;
	}
	public int size(){
		return size;
	}
	public void clear(){
		size=0;
	}
	public int emptyReverseInto(T[] t){
		//Note that if we have only tossed things and didn't hit the max yet, it is not heapified at this point. Then we need to heapify it ourselves
		int itemCount=0;
		if(size!=arr.length)
			heapify();
		for(int i=size-1;i>=0;--i,itemCount++)
			t[i]=pop();
		return itemCount;
	}
	
	public boolean hasRoomLeft(){
		return size<arr.length;
	}
	public T peek(){
		if(hasRoomLeft())
			heapify();
		return arr[0];
	}

	private T pop(){
		T ret=arr[0];
		arr[0]=arr[--size];
		heapify(0);
		return ret;
	}
	private void popAndInsert(T t){
		arr[0]=t;
		heapify(0);
	}
	private void heapify(){
		for(int i=parent(size-1);i>=0;--i)
			heapify(i);
	}
	private void heapify(int i){
		int l=left(i);
		int r=right(i);
		int smallest=i;
//		if(l<size && arr[l].score()<arr[i].score())
		if(l<size && arr[l].compareTo(arr[i])<0)
			smallest=l;
//		if(r<size && arr[r].score()<arr[smallest].score())
		if(r<size && arr[r].compareTo(arr[smallest])<0)
			smallest=r;
		if(smallest!=i){
			swap(i,smallest);
			heapify(smallest);
		}
	}
	private int left(int i)		{	return 2*i+1;}
	private int right(int i)	{	return 2*i+2;}
	private int parent(int i)	{	return (i-1)/2;}
	private void swap(int a,int b){
		T tmp=arr[a];
		arr[a]=arr[b];
		arr[b]=tmp;
	}
	
	public static <T extends MinHeap.HasScore & Comparable<T>> T emptyAndExtractMax(MinHeap<T> minHeap) {
		if(minHeap.size()==0)
			return null;
		T[] a=minHeap.makeArray(minHeap.size());
		minHeap.emptyReverseInto(a);
		return a[0];
	}
	
	public static <T extends MinHeap.HasScore & Comparable<T>> T extractMax(MinHeap<T> minHeap) {
		T max=minHeap.arr[0];
		for(int i=1;i<minHeap.size;++i)
			if(max.compareTo(minHeap.arr[i])<0)
				max=minHeap.arr[i];
		return max;
	}
	
	@SuppressWarnings("unchecked")
	private T[] makeArray(int elems) {
		return (T[]) Array.newInstance(MinHeap.HasScore.class, elems);
	}

	public int getNextTieBreaker(){
		return nextTieBreaker++;
	}
}
